/***
*** ioctl.c  
***
***  General description of this file:
***     Device driver source code for General Standards 16AI64
***     family of analog input boards. This file is part of the Linux
***     driver source distribution for this board.
***     
***
***  Copyrights (c):
***     General Standards Corporation (GSC), 2004-2005
***
***  Author:
***     Evan Hillman (evan@generalstandards.com)
***
***  Support:
***     Primary support for this driver is provided by GSC. 
***
***/

#ifndef __KERNEL__
#define __KERNEL__
#endif

//////////////////////////////////////////////////////////////////////////
// set the following flag to trace debug messages in this module.
#ifdef DEBUG
#define TRACE_LOCAL TRUE
#endif

#include <linux/version.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/pci.h>
#include <linux/mm.h>
#include <linux/vmalloc.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/timer.h>
#include <linux/ioctl.h>

#include <asm/io.h>
#include <asm/system.h>
#include <asm/uaccess.h>

#include "sysdep.h"

#include "gsc16AI64_ioctl.h"
#include "internals.h"
#include "plx_regs.h"

/* ioctl file operation: this does the bulk of the work */
int device_ioctl(struct inode *inode, struct file *fp, unsigned int num, unsigned long arg)
{
  struct device_board *device = (struct device_board *)fp->private_data;
  struct device_register_params regs;
  unsigned long ulval, regval=0;
  //msg("(%d): device pointer: %p\n", device->minor,device);

  /* main ioctl function dispatch */
  /* 'break' at the end of branches which need to wait for the */
  /* channels ready condition, 'return' from others */
  switch (num) {
  case IOCTL_DEVICE_NO_COMMAND:
   return (0);

  case IOCTL_DEVICE_INIT_BOARD:
	  msg("(%d): initializing board\n", device->minor);
#ifdef TRACE_LOCAL
	  //regdump(device);
#endif

    atomic_set(&device->irq_event_pending[EVENT_INIT_COMPLETE],TRUE);
	  /*
	   *	clear out any pending old interrupts.
	   */
//    DisableIrqPlx(device);

	  /* enable PLX interrupt. */

    EnableIrqPlxLocal(device);
	  /* enable local init complete interrupt. */
	  //writelocal(device,IRQ_0_IDLE|IRQ_0_REQUEST, INT_CTRL_REG);
    EnableIrqLocal(device, GROUP_0, ICR_GROUP_0_IDLE);

    regval = readlocal(device,BOARD_CTRL_REG);
    regval |= BCR_INITIALIZE;
    writelocal(device,regval,BOARD_CTRL_REG);

    mark(0xfffffff1);

	  device->timeout=FALSE;
	  device->watchdog_timer.expires=jiffies+device->timeout_seconds*HZ;
	  add_timer(&device->watchdog_timer);

	  writelocal(device,BCR_INITIALIZE, BOARD_CTRL_REG);
	  //msg("init ->plx int %.8X board control %.8X local int %.8X\n",  readl(IntCntrlStat(device)), readlocal(device,BOARD_CTRL_REG),readlocal(device,INT_CTRL_REG));
      wait_event_interruptible(*device->event_queue_waiting[EVENT_INIT_COMPLETE],(!atomic_read(&device->irq_event_pending[EVENT_INIT_COMPLETE])));
	  if (signal_pending(current)) {
		  errmsg("(%d): Signal termination...\n", device->minor);
		  del_timer_sync(&device->watchdog_timer);
		  DisableIrqPlx(device);
		  return (-ERESTARTSYS); 
	  }

	  if (device->timeout)
	  {
		  errmsg("(%d): timeout during board initialize\n", device->minor);
          atomic_set(&device->irq_event_pending[EVENT_INIT_COMPLETE],FALSE);
		  //device->init_pending = FALSE;
		  device->error = DEVICE_IOCTL_TIMEOUT;
		  return (-EIO);
	  }
	  else
		  del_timer_sync(&device->watchdog_timer);

#ifdef TRACE_LOCAL
	  //regdump(device);
#endif
	  //msg("%d): %.8X %.8X %.8X\n",device->minor, readl(IntCntrlStat(device)), readlocal(device,BOARD_CTRL_REG),readl(DMACmdStatus(device)));
	  //msg("plx int %.8X board control %.8X local int %.8X\n",  readl(IntCntrlStat(device)), readlocal(device,BOARD_CTRL_REG),readlocal(device,INT_CTRL_REG));
	  //msg("(%d): ioctl initialize done\n", device->minor);
	  return (0);

  case IOCTL_DEVICE_START_AUTOCAL:
	  //msg("(%d): calibrating board\n", device->minor);
	  //msg("plx int %.8X board control %.8X local int %.8X\n", readl(IntCntrlStat(device)), readlocal(device,BOARD_CTRL_REG),readlocal(device,INT_CTRL_REG));
#ifdef TRACE_LOCAL
	  //regdump(device);
#endif
	  msg(":     base_address[0]=0x%lX, base_address[2]=0x%lX\n",device->pdev->resource[0].start, device->pdev->resource[2].start);

    EnableIrqPlxLocal(device);
    atomic_set(&device->irq_event_pending[EVENT_AUTOCAL_COMPLETE],FALSE);
    EnableIrqLocal(device, GROUP_0, ICR_GROUP_0_AUTOCAL_COMPLETE);

	  /* start autocal */
      atomic_set(&device->irq_event_pending[EVENT_AUTOCAL_COMPLETE],TRUE);
	  /* enable PLX interrupt. */
	  writel(readl(IntCntrlStat(device)) | IRQ_PCI_ENABLE | IRQ_LOCAL_PCI_ENABLE | IRQ_LOCAL_ENABLE, IntCntrlStat(device));
	  writelocal(device,regval, BOARD_CTRL_REG);

#ifdef TRACE_LOCAL
	  msg("(%d): about to wait calibrating board\n", device->minor);
	 // regdump(device);
#endif
	  /* wait for complete */
	  device->timeout=FALSE;
	  device->watchdog_timer.expires=jiffies+device->timeout_seconds*HZ;
	  add_timer(&device->watchdog_timer);

	  regval = readlocal(device,BOARD_CTRL_REG);
	  regval |= BCR_AUTOCAL;
	  writelocal(device,regval, BOARD_CTRL_REG);

      wait_event_interruptible(*device->event_queue_waiting[EVENT_AUTOCAL_COMPLETE],(!atomic_read(&device->irq_event_pending[EVENT_AUTOCAL_COMPLETE])));
	  if (signal_pending(current)) {
		  errmsg("(%d): Signal termination...\n", device->minor);
		  del_timer_sync(&device->watchdog_timer);
		  DisableIrqPlx(device);
		  return (-ERESTARTSYS); 
	  }

	  if (device->timeout) {
		  errmsg("(%d): timeout when calibrating board\n", device->minor);
          atomic_set(&device->irq_event_pending[EVENT_AUTOCAL_COMPLETE],FALSE);
		  device->error = DEVICE_AUTOCAL_FAILED;
		  return (-EIO);
	  }
	  else
		  del_timer_sync(&device->watchdog_timer);

	  //msg(KERN_INFO GSC_NAME "(%d): after wait calibrating board\n", device->minor);

	  return (0);

  case IOCTL_DEVICE_GET_DEVICE_TYPE:
	  put_user_ret(device->board_type, (unsigned long *)arg, (-EFAULT));
	  return (0);

  case IOCTL_DEVICE_READ_REGISTER:
	//msg("(%d): IOCTL_DEVICE_READ_REGISTER\n", device->minor);
    copy_from_user_ret(&regs, (void *)arg, sizeof(regs), (-EFAULT));
    if ((regs.DeviceRegister < BOARD_CTRL_REG) || (regs.DeviceRegister > SCAN_SYNCH_CTRL_REG)) {
      device->error = DEVICE_INVALID_PARAMETER;
      return (-EIO);
    }
//	msg("(%d): IOCTL_DEVICE_READ_REGISTER addr %p register %x\n", device->minor,device->local_addr,regs.DeviceRegister);
    regs.ulValue = readlocal(device, regs.DeviceRegister);
//	msg("(%d): IOCTL_DEVICE_READ_REGISTER before copy\n", device->minor);
    copy_to_user_ret((void *)arg, &regs, sizeof(regs), (-EFAULT));
//	msg("(%d): IOCTL_DEVICE_READ_REGISTER after copy\n", device->minor);
    return (0);
 
  case IOCTL_DEVICE_WRITE_REGISTER:
    copy_from_user_ret(&regs, (void *)arg, sizeof(regs), (-EFAULT));
    if ((regs.DeviceRegister < BOARD_CTRL_REG) || (regs.DeviceRegister > SCAN_SYNCH_CTRL_REG)) {
      device->error = DEVICE_INVALID_PARAMETER;
      return (-EIO);
    }
    writel(regs.ulValue, device->local_addr + regs.DeviceRegister);
    return (0);

  case IOCTL_DEVICE_ENABLE_DMA:
	  get_user_ret(ulval, (unsigned long *)arg, (-EFAULT));
	  if ((ulval < DMA_DISABLE) || (ulval > DMA_DEMAND_MODE)) {
		  device->error = DEVICE_INVALID_PARAMETER;
	  }
	  device->dmaState = ulval;
	  return (0);

  case IOCTL_DEVICE_GET_DEVICE_ERROR:
	  put_user_ret(device->error, (unsigned long *)arg, (-EFAULT));
	  return (0);

  case IOCTL_DEVICE_INPUT_RANGE_CONFIG:
	get_user_ret(ulval, (unsigned long *)arg, (-EFAULT));
    if ((ulval < INPUT_RANGE_2_5) || (ulval > INPUT_RANGE_10)) {
        msg("IOCTL_DEVICE_INPUT_RANGE_CONFIG invalid parameter \n");
      device->error = DEVICE_INVALID_PARAMETER;
      return (-EIO);
    }
	mark(0xfffffff1);
    regval = readlocal(device,BOARD_CTRL_REG);
    regval &= (~(BCR_RANGE_MASK));
    regval |= (ulval<<BCR_RANGE_SHIFT);
    writelocal(device,regval, BOARD_CTRL_REG);
	//return (0);
    break;

  case IOCTL_DEVICE_INPUT_MODE_CONFIG:
	  get_user_ret(ulval, (unsigned long *)arg, (-EFAULT));
	  if ((ulval < INPUT_DIFFERENTIAL) || (ulval > INPUT_VREF_TEST)) {
		  device->error = DEVICE_INVALID_PARAMETER;
		  return (-EIO);
	  }
	  mark(0xfffffff2);
	  regval = readlocal(device,BOARD_CTRL_REG);
	  regval &= (~(INPUT_MODE_MASK));
	  regval |= (ulval<<INPUT_MODE_SHIFT);
	  writelocal(device,regval, BOARD_CTRL_REG);
	  readlocal(device,SCAN_SYNCH_CTRL_REG);

//	  return (0);
	  break;

  case IOCTL_DEVICE_SS_MODE_CONFIG:
	  get_user_ret(ulval, (unsigned long *)arg, (-EFAULT));
	  if ((ulval < SS_SINGLE_ENDED) || (ulval > SS_FULL_DIFFERENTIAL)) {
		  device->error = DEVICE_INVALID_PARAMETER;
		  return (-EIO);
	  }
	  mark(0xfffffff3);
	  regval = readlocal(device,BOARD_CTRL_REG);
	  regval &= (~(SS_MODE_MASK));
	  regval |= (ulval<<SS_MODE_SHIFT);
	  writelocal(device,regval, BOARD_CTRL_REG);
	  return (0);
	  break;

  case IOCTL_DEVICE_INPUT_FORMAT_CONFIG:
	  get_user_ret(ulval, (unsigned long *)arg, (-EFAULT));
	  if ((ulval < FORMAT_TWO_COMPLEMENT) || (ulval > FORMAT_OFFSET_BINARY)) {
		  device->error = DEVICE_INVALID_PARAMETER;
		  return (-EIO);
	  }
	  mark(0xfffffff4);
	  regval = readlocal(device,BOARD_CTRL_REG);
	  regval &= (~(BCR_FORMAT_MASK));
	  regval |= (ulval<<BCR_OFFSET_BINARY_SHIFT);
	  writelocal(device,regval, BOARD_CTRL_REG);
//	  return (0);
	  break;

  case IOCTL_DEVICE_START_SINGLE_SCAN:
	  regval = readlocal(device,BOARD_CTRL_REG);
	  regval |= BCR_INPUT_SYNCH_START;
	  writelocal(device,regval, BOARD_CTRL_REG);
	  return (0);
	  break;

  case IOCTL_DEVICE_SET_SCAN_SIZE:
	  get_user_ret(ulval, (unsigned long *)arg, (-EFAULT));
	  if ((ulval < SCAN_SINGLE_CHAN) || (ulval > SCAN_64_CHAN)) {
		  device->error = DEVICE_INVALID_PARAMETER;
		  return (-EIO);
	  }
	  mark(0xfffffff5);
	  regval = readlocal(device,SCAN_SYNCH_CTRL_REG);
	  regval &= (~(SSCR_CHAN_SIZE_MASK));
	  regval |= (ulval);
	  writelocal(device,regval, SCAN_SYNCH_CTRL_REG);
	  return (0);
	  break;

  case IOCTL_DEVICE_SET_SCAN_CLOCK:
	  get_user_ret(ulval, (unsigned long *)arg, (-EFAULT));
	  if ((ulval < INTERNAL_RATE_A) || (ulval > BCR_INPUT_SYNCH)) {
		  device->error = DEVICE_INVALID_PARAMETER;
		  return (-EIO);
	  }
	  mark(0xfffffff6);
	  regval = readlocal(device,SCAN_SYNCH_CTRL_REG);

	  regval &= (~(SSCR_SCAN_CLOCK_MASK));
	  regval |= (ulval<<SSCR_CLOCK_SHIFT);
	  writelocal(device,regval, SCAN_SYNCH_CTRL_REG);
	  return (0);
	  break;

  case IOCTL_DEVICE_DISABLE_PCI_INTERRUPTS:
	  {
		  int plx_irq_reg;
		  plx_irq_reg = readl(IntCntrlStat(device));
		  plx_irq_reg &= ~PCI_INT_ENABLE;
		  writel(plx_irq_reg, (IntCntrlStat(device)));
	  }
	  return (0);
	  break;

  case IOCTL_DEVICE_SET_B_CLOCK_SOURCE:
	  get_user_ret(ulval, (unsigned long *)arg, (-EFAULT));
	  if ((ulval < MASTER_CLOCK) || (ulval > RATE_A_OUTPUT)) {
		  device->error = DEVICE_INVALID_PARAMETER;
		  return (-EIO);
	  }
	  regval = readlocal(device,SCAN_SYNCH_CTRL_REG);

	  regval &= (~(SSCR_RATE_B_CLOCK_SOURCE_MASK));
	  regval |= (ulval<<SSCR_RATE_B_CLOCK_SOURCE_SHIFT);
	  writelocal(device,regval, SCAN_SYNCH_CTRL_REG);
	  return (0);
	  break;

  case IOCTL_DEVICE_SINGLE_CHAN_SELECT:
	  get_user_ret(ulval, (unsigned long *)arg, (-EFAULT));
	  if ((ulval < 0) || (ulval > 63)) {
		  device->error = DEVICE_INVALID_PARAMETER;
		  return (-EIO);
	  }
	  regval = readlocal(device,SCAN_SYNCH_CTRL_REG);

	  regval &= (~(SSCR_CHAN_SIZE_MASK));
	  regval &= (~(SSCR_SINGLE_CHANNEL_MASK));
	  regval |= (ulval<<SSCR_SINGLE_CHANNEL_SHIFT);
	  writelocal(device,regval, SCAN_SYNCH_CTRL_REG);
	  return (0);
	  break;

  case IOCTL_DEVICE_SET_SCAN_RATE_A:
	  get_user_ret(ulval, (unsigned long *)arg, (-EFAULT));
	  if ((ulval < 0) || (ulval > 0xFFFF)) {
		  device->error = DEVICE_INVALID_PARAMETER;
		  return (-EIO);
	  }
	  mark(0xfffffff7);
	  regval = readlocal(device,RATE_A_GENERATOR_REG);

	  regval &= (~(RGR_NRATE_MASK));
	  regval |= (ulval);
	  writelocal(device,regval, RATE_A_GENERATOR_REG);
	  return (0);
	  break;

  case IOCTL_DEVICE_SET_SCAN_RATE_B:
	  get_user_ret(ulval, (unsigned long *)arg, (-EFAULT));
	  if ((ulval < 0) || (ulval > 0xFFFF)) {
		  device->error = DEVICE_INVALID_PARAMETER;
		  return (-EIO);
	  }
	  mark(0xfffffff8);
	  regval = readlocal(device,RATE_B_GENERATOR_REG);

	  regval &= (~(RGR_NRATE_MASK));
	  regval |= (ulval);
	  writelocal(device,regval, RATE_B_GENERATOR_REG);
	  return (0);
	  break;

  case IOCTL_DEVICE_GEN_DISABLE_A:
	  regval = readlocal(device,RATE_A_GENERATOR_REG);
	  regval |= RGR_GEN_DISABLE;
	  writelocal(device,regval, RATE_A_GENERATOR_REG);
	  return (0);
	  break;

  case IOCTL_DEVICE_GEN_DISABLE_B:
	  regval = readlocal(device,RATE_B_GENERATOR_REG);
	  regval |= RGR_GEN_DISABLE;
	  writelocal(device,regval, RATE_B_GENERATOR_REG);
	  return (0);
	  break;

  case IOCTL_DEVICE_GEN_ENABLE_A:
	  mark(0xfffffff9);
	  regval = readlocal(device,RATE_A_GENERATOR_REG);
	  regval &= ~(RGR_GEN_DISABLE);
	  writelocal(device,regval, RATE_A_GENERATOR_REG);
	  return (0);
	  break;

  case IOCTL_DEVICE_GEN_ENABLE_B:
	  regval = readlocal(device,RATE_B_GENERATOR_REG);
	  regval &= ~(RGR_GEN_DISABLE);
	  writelocal(device,regval, RATE_B_GENERATOR_REG);
	  return (0);
	  break;

  case IOCTL_DEVICE_SET_BUFFER_THRESHOLD:
	  get_user_ret(ulval, (unsigned long *)arg, (-EFAULT));
	  if ((ulval < 0) || (ulval > 0xFFFF)) {
		  device->error = DEVICE_INVALID_PARAMETER;
		  return (-EIO);
	  }

	  mark(0xfffffffa);
	  regval = readlocal(device,INPUT_BUFF_CTRL_REG);
	  regval &= (~(0xFFFF));
	  regval |= (ulval);
	  device->bufferThreshold = regval;
	  writelocal(device,regval, INPUT_BUFF_CTRL_REG);
	  return (0);
	  break;

  case IOCTL_DEVICE_CLEAR_BUFFER:
	  regval = readlocal(device,INPUT_BUFF_CTRL_REG);
	  regval |= IDC_INPUT_CLEAR_BUFFER;
	  writelocal(device,regval, INPUT_BUFF_CTRL_REG);
	  return (0);
	  break;

  case IOCTL_DEVICE_ENABLE_BUFFER_SS:
	  regval = readlocal(device,INPUT_BUFF_CTRL_REG);
	  regval &= ~(IDC_DISABLE_BUFFER_SS);
	  writelocal(device,regval, INPUT_BUFF_CTRL_REG);
	  return (0);
	  break;

  case IOCTL_DEVICE_DISABLE_BUFFER_SS:
	  regval = readlocal(device,INPUT_BUFF_CTRL_REG);
	  regval |= IDC_DISABLE_BUFFER_SS;
	  writelocal(device,regval, INPUT_BUFF_CTRL_REG);
	  return (0);
	  break;
          
  case IOCTL_DEVICE_ENABLE_DMA_ALMOST_SS:
	  regval = readlocal(device,INPUT_BUFF_CTRL_REG);
	  regval |= IDC_TERMINATE_DMA_ALMOST_EMPTY_SS;
	  writelocal(device,regval, INPUT_BUFF_CTRL_REG);
	  return (0);
	  break;

  case IOCTL_DEVICE_DISABLE_DMA_ALMOST_SS:
	  regval = readlocal(device,INPUT_BUFF_CTRL_REG);
	  regval &= ~(IDC_TERMINATE_DMA_ALMOST_EMPTY_SS);
	  writelocal(device,regval, INPUT_BUFF_CTRL_REG);
	  return (0);
	  break;

  case IOCTL_DEVICE_SET_TIMEOUT:
	  get_user_ret(ulval, (unsigned long *)arg, (-EFAULT));
	  if ((ulval < 0) || (ulval > 0xffff)) {
		  device->error = DEVICE_INVALID_PARAMETER;
		  return (-EIO);
	  }
	  device->timeout_seconds=ulval;
	  return (0);
	  break;

  case IOCTL_DEVICE_FILL_BUFFER:
	  get_user_ret(ulval, (int *)arg, (-EFAULT));
	  ulval = ulval ?	1 :	0;
	  device->fillBuffer = ulval;
	  return (0);
      break;

  case IOCTL_DEVICE_DETECT_OVERFLOW:
	  get_user_ret(ulval, (int *)arg, (-EFAULT));
	  ulval = ulval ?	1 :	0;
	  device->detectOverflow = ulval;
	  return (0);
      break;

  case IOCTL_DEVICE_SET_UNIPOLAR_INPUTS_SSA:   
      get_user_ret(ulval, (unsigned long *)arg, (-EFAULT));
      if ((ulval != BIPOLAR_INPUTS) && (ulval != UNIPOLAR_INPUTS)) {
          device->error = DEVICE_INVALID_PARAMETER;
          return (-EIO);
      }
      regval = readlocal(device,BOARD_CTRL_REG);
      regval &= (~(BCR_UNIPOLAR_INPUTS_SSA));
      regval |= (ulval);
      writelocal(device,regval, BOARD_CTRL_REG);
      break;

  case IOCTL_DEVICE_SET_DIFF_PROCESSING_SSA:   
      get_user_ret(ulval, (unsigned long *)arg, (-EFAULT));
      if ((ulval < SINGLE_ENDED_PROC) || (ulval > FULL_DIFFERENTIAL_PROC)) {
          device->error = DEVICE_INVALID_PARAMETER;
          return (-EIO);
      }
      regval = readlocal(device,BOARD_CTRL_REG);
      regval &= (~(BCR_DIFF_PROC_MASK_SSA));
      regval |= (ulval<<BCR_DIFF_PROC_SHIFT_SSA);
      writelocal(device,regval, BOARD_CTRL_REG);
      break;

  case IOCTL_DEVICE_READ_BUFFER_UNDERFLOW_SSA: 
      regval = readlocal(device,BOARD_CTRL_REG);
      regval &= BCR_BUFFER_UNDERFLOW_SSA;
      put_user_ret(regval, (unsigned long *)arg, (-EFAULT));
     break;

  case IOCTL_DEVICE_READ_BUFFER_OVERFLOW_SSA:  
      regval = readlocal(device,BOARD_CTRL_REG);
      regval &= BCR_BUFFER_OVERFLOW_SSA;
      put_user_ret(regval, (unsigned long *)arg, (-EFAULT));
      break;

  case IOCTL_DEVICE_SET_SCAN_MARKING_SSA:      
      get_user_ret(ulval, (unsigned long *)arg, (-EFAULT));

      regval = readlocal(device,BOARD_CTRL_REG);
      if (ulval)
      {
          regval |= (ENABLE_SCAN_MARKING_SSA);
      }
      else
      {
          regval &= (~(ENABLE_SCAN_MARKING_SSA));
      }
      writelocal(device,regval, BOARD_CTRL_REG);
      break;

  case IOCTL_DEVICE_SET_DEMAND_MODE_DISABLE_SSA:  
      get_user_ret(ulval, (unsigned long *)arg, (-EFAULT));

      regval = readlocal(device,BOARD_CTRL_REG);
      if (ulval)
      {
          regval |= (DISABLE_DEMAND_MODE_SSA);
      }
      else
      {
          regval &= (~(DISABLE_DEMAND_MODE_SSA));
      }
      writelocal(device,regval, BOARD_CTRL_REG);
      break;

  case IOCTL_DEVICE_SET_THRESHOLD_4X_SSA:      
      get_user_ret(ulval, (unsigned long *)arg, (-EFAULT));
      regval = readlocal(device,INPUT_BUFF_CTRL_REG);
      if (ulval)
      {
          regval |= (IDC_THRESHOLD_4X_SSA);
      }
      else
      {
          regval &= (~(IDC_THRESHOLD_4X_SSA));
      }
      writelocal(device,regval, INPUT_BUFF_CTRL_REG);
      break;

  case IOCTL_DEVICE_RESET_BUFFER_UNDERFLOW_SSA: 
      regval = readlocal(device,BOARD_CTRL_REG);
      regval &= ~BCR_BUFFER_UNDERFLOW_SSA;
      writelocal(device,regval, BOARD_CTRL_REG);
      break;

  case IOCTL_DEVICE_RESET_BUFFER_OVERFLOW_SSA: 
      regval = readlocal(device,BOARD_CTRL_REG);
      regval &= ~BCR_BUFFER_OVERFLOW_SSA;
      writelocal(device,regval, BOARD_CTRL_REG);
      break;

  case IOCTL_DEVICE_SET_AUX_0_MODE_SSA:      
      get_user_ret(ulval, (unsigned long *)arg, (-EFAULT));
      if ((ulval < AUX_INACTIVE) || (ulval > AUX_ACTIVE_OUT)) {
          device->error = DEVICE_INVALID_PARAMETER;
          return (-EIO);
      }
      regval = readlocal(device,AUX_SYNCH_IO_CTRL);
      regval &= (~(ASC_MODE_MASK<<ASC_CONTROL_MODE_0_SHIFT));
      regval |= (ulval<<ASC_CONTROL_MODE_0_SHIFT);
      writelocal(device,regval, AUX_SYNCH_IO_CTRL);
      break;

  case IOCTL_DEVICE_SET_AUX_1_MODE_SSA:      
      get_user_ret(ulval, (unsigned long *)arg, (-EFAULT));
      if ((ulval < AUX_INACTIVE) || (ulval > AUX_ACTIVE_OUT)) {
          device->error = DEVICE_INVALID_PARAMETER;
          return (-EIO);
      }
      regval = readlocal(device,AUX_SYNCH_IO_CTRL);
      regval &= (~(ASC_MODE_MASK<<ASC_CONTROL_MODE_1_SHIFT));
      regval |= (ulval<<ASC_CONTROL_MODE_1_SHIFT);
      writelocal(device,regval, AUX_SYNCH_IO_CTRL);
      break;

  case IOCTL_DEVICE_SET_AUX_2_MODE_SSA:      
      get_user_ret(ulval, (unsigned long *)arg, (-EFAULT));
      if ((ulval < AUX_INACTIVE) || (ulval > AUX_ACTIVE_OUT)) {
          device->error = DEVICE_INVALID_PARAMETER;
          return (-EIO);
      }
      regval = readlocal(device,AUX_SYNCH_IO_CTRL);
      regval &= (~(ASC_MODE_MASK<<ASC_CONTROL_MODE_2_SHIFT));
      regval |= (ulval<<ASC_CONTROL_MODE_2_SHIFT);
      writelocal(device,regval, AUX_SYNCH_IO_CTRL);
      break;

  case IOCTL_DEVICE_SET_AUX_3_MODE_SSA:      
      get_user_ret(ulval, (unsigned long *)arg, (-EFAULT));
      if ((ulval < AUX_INACTIVE) || (ulval > AUX_ACTIVE_OUT)) {
          device->error = DEVICE_INVALID_PARAMETER;
          return (-EIO);
      }
      regval = readlocal(device,AUX_SYNCH_IO_CTRL);
      regval &= (~(ASC_MODE_MASK<<ASC_CONTROL_MODE_3_SHIFT));
      regval |= (ulval<<ASC_CONTROL_MODE_3_SHIFT);
      writelocal(device,regval, AUX_SYNCH_IO_CTRL);
      break;

  case IOCTL_DEVICE_SET_INPUT_INVERT_MODE_SSA:      
      get_user_ret(ulval, (unsigned long *)arg, (-EFAULT));
      regval = readlocal(device,AUX_SYNCH_IO_CTRL);
      if (ulval)
      {
          regval |= (ASC_INVERT_INPUTS);
      }
      else
      {
          regval &= (~(ASC_INVERT_INPUTS));
      }
      writelocal(device,regval, AUX_SYNCH_IO_CTRL);
      break;

  case IOCTL_DEVICE_SET_OUTPUT_INVERT_MODE_SSA:      
      get_user_ret(ulval, (unsigned long *)arg, (-EFAULT));
      regval = readlocal(device,AUX_SYNCH_IO_CTRL);
      if (ulval)
      {
          regval |= (ASC_INVERT_OUTPUTS);
      }
      else
      {
          regval &= (~(ASC_INVERT_OUTPUTS));
      }
      writelocal(device,regval, AUX_SYNCH_IO_CTRL);
      break;

  case IOCTL_DEVICE_SET_NOISE_SUPPRESSION_SSA:      
      get_user_ret(ulval, (unsigned long *)arg, (-EFAULT));
      regval = readlocal(device,AUX_SYNCH_IO_CTRL);
      if (ulval)
      {
          regval |= (ASC_NOISE_SUPPRESS);
      }
      else
      {
          regval &= (~(ASC_NOISE_SUPPRESS));
      }
      writelocal(device,regval, AUX_SYNCH_IO_CTRL);
      break;

  case IOCTL_DEVICE_SET_HW_BUFFER_SIZE:
      get_user_ret(ulval, (unsigned long *)arg, (-EFAULT));
      if ((ulval < 0) || (ulval > MAX_BUFFER_SIZE)) {
          device->error = DEVICE_INVALID_PARAMETER;
          return (-EIO);
      }
      device->hwBufferSize=ulval;
      msg("IOCTL_DEVICE_SET_HW_BUFFER_SIZE buffer size set to %.8X",device->hwBufferSize);
      return (0);
      break;

      
  default:
	errmsg("(%d): ERROR - unexpected IOCTL\n", device->minor);
    return (-EINVAL);
  } // switch
  return (0);
} // device_ioctl()
